@@ -1,3 +1,4 @@
+/* ex: set tabstop=2 shiftwidth=2 expandtab: */
 /* GStreamer
  *
  * This library is free software; you can redistribute it and/or
@@ -27,6 +28,7 @@
 #include "gstrtph264pay.h"
 
 
+#define IDR_TYPE_ID  5
 #define SPS_TYPE_ID  7
 #define PPS_TYPE_ID  8
 #define USE_MEMCMP
@@ -90,6 +92,8 @@ gst_h264_scan_mode_get_type (void)
 #define DEFAULT_PROFILE_LEVEL_ID        NULL
 #define DEFAULT_SPROP_PARAMETER_SETS    NULL
 #define DEFAULT_SCAN_MODE               GST_H264_SCAN_MODE_MULTI_NAL
+#define DEFAULT_BUFFER_LIST             FALSE
+#define DEFAULT_CONFIG_INTERVAL		      0
 
 enum
 {
@@ -97,6 +101,8 @@ enum
   PROP_PROFILE_LEVEL_ID,
   PROP_SPROP_PARAMETER_SETS,
   PROP_SCAN_MODE,
+  PROP_BUFFER_LIST,
+  PROP_CONFIG_INTERVAL,
   PROP_LAST
 };
 
@@ -145,8 +151,7 @@ gst_rtp_h264_pay_class_init (GstRtpH264PayClass * klass)
   g_object_class_install_property (G_OBJECT_CLASS (klass),
       PROP_PROFILE_LEVEL_ID, g_param_spec_string ("profile-level-id",
           "profile-level-id",
-          "The base64 profile-level-id to set in out caps (set to NULL to "
-          "extract from stream)",
+          "The base64 profile-level-id to set in the sink caps (deprecated)",
           DEFAULT_PROFILE_LEVEL_ID,
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
@@ -165,6 +170,21 @@ gst_rtp_h264_pay_class_init (GstRtpH264PayClass * klass)
           GST_TYPE_H264_SCAN_MODE, DEFAULT_SCAN_MODE,
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
+  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_BUFFER_LIST,
+      g_param_spec_boolean ("buffer-list", "Buffer List",
+          "Use Buffer Lists",
+          DEFAULT_BUFFER_LIST, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
+  g_object_class_install_property (G_OBJECT_CLASS (klass),
+      PROP_CONFIG_INTERVAL,
+      g_param_spec_uint ("config-interval",
+          "SPS PPS Send Interval",
+          "Send SPS and PPS Insertion Interval in seconds (sprop parameter sets "
+          "will be multiplexed in the data stream when detected.) (0 = disabled)",
+          0, 3600, DEFAULT_CONFIG_INTERVAL,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)
+      );
+
   gobject_class->finalize = gst_rtp_h264_pay_finalize;
 
   gstbasertppayload_class->set_caps = gst_rtp_h264_pay_setcaps;
@@ -177,10 +197,25 @@ gst_rtp_h264_pay_class_init (GstRtpH264PayClass * klass)
 static void
 gst_rtp_h264_pay_init (GstRtpH264Pay * rtph264pay, GstRtpH264PayClass * klass)
 {
+  rtph264pay->queue = g_array_new (FALSE, FALSE, sizeof (guint));
   rtph264pay->profile = 0;
   rtph264pay->sps = NULL;
   rtph264pay->pps = NULL;
+  rtph264pay->last_spspps = -1;
   rtph264pay->scan_mode = GST_H264_SCAN_MODE_MULTI_NAL;
+  rtph264pay->buffer_list = DEFAULT_BUFFER_LIST;
+  rtph264pay->spspps_interval = DEFAULT_CONFIG_INTERVAL;
+}
+
+static void
+gst_rtp_h264_pay_clear_sps_pps (GstRtpH264Pay * rtph264pay)
+{
+  g_list_foreach (rtph264pay->sps, (GFunc) gst_mini_object_unref, NULL);
+  g_list_free (rtph264pay->sps);
+  rtph264pay->sps = NULL;
+  g_list_foreach (rtph264pay->pps, (GFunc) gst_mini_object_unref, NULL);
+  g_list_free (rtph264pay->pps);
+  rtph264pay->pps = NULL;
 }
 
 static void
@@ -190,15 +225,62 @@ gst_rtp_h264_pay_finalize (GObject * object)
 
   rtph264pay = GST_RTP_H264_PAY (object);
 
-  g_free (rtph264pay->sps);
-  g_free (rtph264pay->pps);
+  g_array_free (rtph264pay->queue, TRUE);
+
+  gst_rtp_h264_pay_clear_sps_pps (rtph264pay);
 
-  g_free (rtph264pay->profile_level_id);
   g_free (rtph264pay->sprop_parameter_sets);
 
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+/* take the currently configured SPS and PPS lists and set them on the caps as
+ * sprop-parameter-sets */
+static gboolean
+gst_rtp_h264_pay_set_sps_pps (GstBaseRTPPayload * basepayload)
+{
+  GstRtpH264Pay *payloader = GST_RTP_H264_PAY (basepayload);
+  gchar *profile;
+  gchar *set;
+  GList *walk;
+  GString *sprops;
+  guint count;
+  gboolean res;
+
+  sprops = g_string_new ("");
+  count = 0;
+
+  /* build the sprop-parameter-sets */
+  for (walk = payloader->sps; walk; walk = g_list_next (walk)) {
+    GstBuffer *sps_buf = GST_BUFFER_CAST (walk->data);
+
+    set =
+        g_base64_encode (GST_BUFFER_DATA (sps_buf), GST_BUFFER_SIZE (sps_buf));
+    g_string_append_printf (sprops, "%s%s", count ? "," : "", set);
+    g_free (set);
+    count++;
+  }
+  for (walk = payloader->pps; walk; walk = g_list_next (walk)) {
+    GstBuffer *pps_buf = GST_BUFFER_CAST (walk->data);
+
+    set =
+        g_base64_encode (GST_BUFFER_DATA (pps_buf), GST_BUFFER_SIZE (pps_buf));
+    g_string_append_printf (sprops, "%s%s", count ? "," : "", set);
+    g_free (set);
+    count++;
+  }
+
+  /* profile is 24 bit. Force it to respect the limit */
+  profile = g_strdup_printf ("%06x", payloader->profile & 0xffffff);
+  /* combine into output caps */
+  res = gst_basertppayload_set_outcaps (basepayload,
+      "sprop-parameter-sets", G_TYPE_STRING, sprops->str, NULL);
+  g_string_free (sprops, TRUE);
+  g_free (profile);
+
+  return res;
+}
+
 static gboolean
 gst_rtp_h264_pay_setcaps (GstBaseRTPPayload * basepayload, GstCaps * caps)
 {
@@ -219,11 +301,8 @@ gst_rtp_h264_pay_setcaps (GstBaseRTPPayload * basepayload, GstCaps * caps)
   /* packetized AVC video has a codec_data */
   if ((value = gst_structure_get_value (str, "codec_data"))) {
     GstBuffer *buffer;
-    GString *sprops;
     guint num_sps, num_pps;
-    gint i, count, nal_size;
-    gint profile;
-    gchar *profile_str;
+    gint i, nal_size;
 
     GST_DEBUG_OBJECT (rtph264pay, "have packetized h264");
     rtph264pay->packetized = TRUE;
@@ -242,8 +321,8 @@ gst_rtp_h264_pay_setcaps (GstBaseRTPPayload * basepayload, GstCaps * caps)
     /* AVCProfileIndication */
     /* profile_compat */
     /* AVCLevelIndication */
-    profile = (data[1] << 16) | (data[2] << 8) | data[3];
-    GST_DEBUG_OBJECT (rtph264pay, "profile %06x", profile);
+    rtph264pay->profile = (data[1] << 16) | (data[2] << 8) | data[3];
+    GST_DEBUG_OBJECT (rtph264pay, "profile %06x", rtph264pay->profile);
 
     /* 6 bits reserved | 2 bits lengthSizeMinusOne */
     /* this is the number of bytes in front of the NAL units to mark their
@@ -258,11 +337,8 @@ gst_rtp_h264_pay_setcaps (GstBaseRTPPayload * basepayload, GstCaps * caps)
     size -= 6;
 
     /* create the sprop-parameter-sets */
-    sprops = g_string_new ("");
-    count = 0;
-
     for (i = 0; i < num_sps; i++) {
-      gchar *set;
+      GstBuffer *sps_buf;
 
       if (size < 2)
         goto avcc_error;
@@ -276,10 +352,10 @@ gst_rtp_h264_pay_setcaps (GstBaseRTPPayload * basepayload, GstCaps * caps)
       if (size < nal_size)
         goto avcc_error;
 
-      set = g_base64_encode (data, nal_size);
-      g_string_append_printf (sprops, "%s%s", count ? "," : "", set);
-      count++;
-      g_free (set);
+      /* make a buffer out of it and add to SPS list */
+      sps_buf = gst_buffer_new_and_alloc (nal_size);
+      memcpy (GST_BUFFER_DATA (sps_buf), data, nal_size);
+      rtph264pay->sps = g_list_append (rtph264pay->sps, sps_buf);
 
       data += nal_size;
       size -= nal_size;
@@ -294,7 +370,7 @@ gst_rtp_h264_pay_setcaps (GstBaseRTPPayload * basepayload, GstCaps * caps)
 
     GST_DEBUG_OBJECT (rtph264pay, "num PPS %u", num_pps);
     for (i = 0; i < num_pps; i++) {
-      gchar *set;
+      GstBuffer *pps_buf;
 
       if (size < 2)
         goto avcc_error;
@@ -308,23 +384,17 @@ gst_rtp_h264_pay_setcaps (GstBaseRTPPayload * basepayload, GstCaps * caps)
       if (size < nal_size)
         goto avcc_error;
 
-      set = g_base64_encode (data, nal_size);
-      g_string_append_printf (sprops, "%s%s", count ? "," : "", set);
-      count++;
-      g_free (set);
+      /* make a buffer out of it and add to PPS list */
+      pps_buf = gst_buffer_new_and_alloc (nal_size);
+      memcpy (GST_BUFFER_DATA (pps_buf), data, nal_size);
+      rtph264pay->pps = g_list_append (rtph264pay->pps, pps_buf);
 
       data += nal_size;
       size -= nal_size;
     }
-    GST_DEBUG_OBJECT (rtph264pay, "sprops %s", sprops->str);
-
-    profile_str = g_strdup_printf ("%06x", profile);
-    gst_basertppayload_set_outcaps (basepayload, "profile-level-id",
-        G_TYPE_STRING, profile_str,
-        "sprop-parameter-sets", G_TYPE_STRING, sprops->str, NULL);
-    g_free (profile_str);
-
-    g_string_free (sprops, TRUE);
+    /* and update the caps with the collected data */
+    if (!gst_rtp_h264_pay_set_sps_pps (basepayload))
+      return FALSE;
   } else {
     GST_DEBUG_OBJECT (rtph264pay, "have bytestream h264");
     rtph264pay->packetized = FALSE;
@@ -390,56 +460,18 @@ next_start_code (guint8 * data, guint size)
   return size;
 }
 
-/* we don't use memcpy but this faster version (around 20%) because we need to
- * perform it on all data. */
 static gboolean
-is_nal_equal (const guint8 * nal1, const guint8 * nal2, guint len)
-{
-  /* if we have a 64-bit processor, we may use guint64 to make 
-   * this go faster. Otherwise with 32 bits, we are already
-   * going faster than byte to byte compare.
-   */
-  guint remainder = len & 0x3;
-  guint num_int = len >> 2;
-  guint32 *pu1 = (guint32 *) nal1, *pu2 = (guint32 *) nal2;
-  guint i;
-
-  /* compare by groups of sizeof(guint32) bytes */
-  for (i = 0; i < num_int; i++) {
-    /* XOR is faster than CMP?... */
-    if (pu1[i] ^ pu2[i])
-      return FALSE;
-  }
-
-  /* check that the remaining bytes are still equal */
-  if (!remainder) {
-    return TRUE;
-  } else if (1 == remainder) {
-    --len;
-    return (nal1[len] == nal2[len]);
-  } else {                      /* 2 or 3 */
-    if (remainder & 1) {        /* -1 if 3 bytes left */
-      --len;
-      if (nal1[len] != nal2[len])
-        return FALSE;
-    }
-    /* last 2 bytes */
-    return ((nal1[len - 1] == nal2[len - 1])    /* -1 */
-        &&(nal1[len - 2] == nal2[len - 2]));    /* -2 */
-  }
-}
-
-static void
 gst_rtp_h264_pay_decode_nal (GstRtpH264Pay * payloader,
-    guint8 * data, guint size, gboolean * updated)
+    guint8 * data, guint size, GstClockTime timestamp)
 {
   guint8 *sps = NULL, *pps = NULL;
   guint sps_len = 0, pps_len = 0;
   guint8 header, type;
   guint len;
+  gboolean updated;
 
   /* default is no update */
-  *updated = FALSE;
+  updated = FALSE;
 
   GST_DEBUG ("NAL payload len=%u", size);
 
@@ -448,7 +480,8 @@ gst_rtp_h264_pay_decode_nal (GstRtpH264Pay * payloader,
   type = header & 0x1f;
 
   /* keep sps & pps separately so that we can update either one 
-   * independently */
+   * independently. We also record the timestamp of the last SPS/PPS so 
+   * that we can insert them at regular intervals and when needed. */
   if (SPS_TYPE_ID == type) {
     /* encode the entire SPS NAL in base64 */
     GST_DEBUG ("Found SPS %x %x %x Len=%u", (header >> 7),
@@ -456,6 +489,9 @@ gst_rtp_h264_pay_decode_nal (GstRtpH264Pay * payloader,
 
     sps = data;
     sps_len = len;
+    /* remember when we last saw SPS */
+    if (timestamp != -1)
+      payloader->last_spspps = timestamp;
   } else if (PPS_TYPE_ID == type) {
     /* encoder the entire PPS NAL in base64 */
     GST_DEBUG ("Found PPS %x %x %x Len = %u",
@@ -463,94 +499,129 @@ gst_rtp_h264_pay_decode_nal (GstRtpH264Pay * payloader,
 
     pps = data;
     pps_len = len;
+    /* remember when we last saw PPS */
+    if (timestamp != -1)
+      payloader->last_spspps = timestamp;
   } else {
     GST_DEBUG ("NAL: %x %x %x Len = %u", (header >> 7),
         (header >> 5) & 3, type, len);
   }
 
-
   /* If we encountered an SPS and/or a PPS, check if it's the
    * same as the one we have. If not, update our version and
-   * set *updated to TRUE
+   * set updated to TRUE
    */
   if (sps_len > 0) {
-    if ((payloader->sps_len != sps_len)
-        || !is_nal_equal (payloader->sps, sps, sps_len)) {
-      payloader->profile = (sps[1] << 16) + (sps[2] << 8) + sps[3];
+    GstBuffer *sps_buf;
 
-      GST_DEBUG ("Profile level IDC = %06x", payloader->profile);
+    if (payloader->sps != NULL) {
+      sps_buf = GST_BUFFER_CAST (payloader->sps->data);
 
-      if (payloader->sps_len)
-        g_free (payloader->sps);
+      if ((GST_BUFFER_SIZE (sps_buf) != sps_len)
+          || memcmp (GST_BUFFER_DATA (sps_buf), sps, sps_len)) {
+        /* something changed, update */
+        payloader->profile = (sps[1] << 16) + (sps[2] << 8) + sps[3];
+        GST_DEBUG ("Profile level IDC = %06x", payloader->profile);
+        updated = TRUE;
+      }
+    } else {
+      /* no previous SPS, update */
+      updated = TRUE;
+    }
 
-      payloader->sps = sps_len ? g_new (guint8, sps_len) : NULL;
-      memcpy (payloader->sps, sps, sps_len);
-      payloader->sps_len = sps_len;
-      *updated = TRUE;
+    if (updated) {
+      sps_buf = gst_buffer_new_and_alloc (sps_len);
+      memcpy (GST_BUFFER_DATA (sps_buf), sps, sps_len);
+
+      if (payloader->sps) {
+        /* replace old buffer */
+        gst_buffer_unref (payloader->sps->data);
+        payloader->sps->data = sps_buf;
+      } else {
+        /* add new buffer */
+        payloader->sps = g_list_prepend (payloader->sps, sps_buf);
+      }
     }
   }
 
   if (pps_len > 0) {
-    if ((payloader->pps_len != pps_len)
-        || !is_nal_equal (payloader->pps, pps, pps_len)) {
-      if (payloader->pps_len)
-        g_free (payloader->pps);
+    GstBuffer *pps_buf;
 
-      payloader->pps = pps_len ? g_new (guint8, pps_len) : NULL;
-      memcpy (payloader->pps, pps, pps_len);
-      payloader->pps_len = pps_len;
-      *updated = TRUE;
+    if (payloader->pps != NULL) {
+      pps_buf = GST_BUFFER_CAST (payloader->pps->data);
+
+      if ((GST_BUFFER_SIZE (pps_buf) != pps_len)
+          || memcmp (GST_BUFFER_DATA (pps_buf), pps, pps_len)) {
+        /* something changed, update */
+        updated = TRUE;
+      }
+    } else {
+      /* no previous SPS, update */
+      updated = TRUE;
+    }
+
+    if (updated) {
+      pps_buf = gst_buffer_new_and_alloc (pps_len);
+      memcpy (GST_BUFFER_DATA (pps_buf), pps, pps_len);
+
+      if (payloader->pps) {
+        /* replace old buffer */
+        gst_buffer_unref (payloader->pps->data);
+        payloader->pps->data = pps_buf;
+      } else {
+        /* add new buffer */
+        payloader->pps = g_list_prepend (payloader->pps, pps_buf);
+      }
     }
   }
-}
-
-static void
-gst_rtp_h264_pay_parse_sps_pps (GstBaseRTPPayload * basepayload,
-    guint8 * data, guint size)
-{
-  gboolean update = FALSE;
-  GstRtpH264Pay *payloader = GST_RTP_H264_PAY (basepayload);
-
-  gst_rtp_h264_pay_decode_nal (payloader, data, size, &update);
-
-  /* if has new SPS & PPS, update the output caps */
-  if (update) {
-    gchar *profile;
-    gchar *sps;
-    gchar *pps;
-    gchar *sprops;
-
-    /* profile is 24 bit. Force it to respect the limit */
-    profile = g_strdup_printf ("%06x", payloader->profile & 0xffffff);
-
-    /* build the sprop-parameter-sets */
-    sps = (payloader->sps_len > 0)
-        ? g_base64_encode (payloader->sps, payloader->sps_len) : NULL;
-    pps = (payloader->pps_len > 0)
-        ? g_base64_encode (payloader->pps, payloader->pps_len) : NULL;
-
-    if (sps)
-      sprops = g_strjoin (",", sps, pps, NULL);
-    else
-      sprops = g_strdup (pps);
-
-    gst_basertppayload_set_outcaps (basepayload, "profile-level-id",
-        G_TYPE_STRING, profile,
-        "sprop-parameter-sets", G_TYPE_STRING, sprops, NULL);
-
-    GST_DEBUG ("outcaps udpate: profile=%s, sps=%s, pps=%s",
-        profile, GST_STR_NULL (sps), GST_STR_NULL (pps));
-
-    g_free (sprops);
-    g_free (profile);
-    g_free (sps);
-    g_free (pps);
-  }
+  return updated;
 }
 
 static GstFlowReturn
 gst_rtp_h264_pay_payload_nal (GstBaseRTPPayload * basepayload, guint8 * data,
-    guint size, GstClockTime timestamp)
+    guint size, GstClockTime timestamp, GstBuffer * buffer_orig);
+
+static GstFlowReturn
+gst_rtp_h264_pay_send_sps_pps (GstBaseRTPPayload * basepayload,
+    GstRtpH264Pay * rtph264pay, GstClockTime timestamp)
+{
+  GstFlowReturn ret = GST_FLOW_OK;
+  GList *walk;
+
+  for (walk = rtph264pay->sps; walk; walk = g_list_next (walk)) {
+    GstBuffer *sps_buf = GST_BUFFER_CAST (walk->data);
+
+    GST_DEBUG_OBJECT (rtph264pay, "inserting SPS in the stream");
+    /* resend SPS */
+    ret = gst_rtp_h264_pay_payload_nal (basepayload,
+        GST_BUFFER_DATA (sps_buf), GST_BUFFER_SIZE (sps_buf), timestamp,
+        sps_buf);
+    /* Not critical here; but throw a warning */
+    if (ret != GST_FLOW_OK)
+      GST_WARNING ("Problem pushing SPS");
+  }
+  for (walk = rtph264pay->pps; walk; walk = g_list_next (walk)) {
+    GstBuffer *pps_buf = GST_BUFFER_CAST (walk->data);
+
+    GST_DEBUG_OBJECT (rtph264pay, "inserting PPS in the stream");
+    /* resend PPS */
+    ret = gst_rtp_h264_pay_payload_nal (basepayload,
+        GST_BUFFER_DATA (pps_buf), GST_BUFFER_SIZE (pps_buf), timestamp,
+        pps_buf);
+    /* Not critical here; but throw a warning */
+    if (ret != GST_FLOW_OK)
+      GST_WARNING ("Problem pushing PPS");
+  }
+
+  if (timestamp != -1)
+    rtph264pay->last_spspps = timestamp;
+
+  return ret;
+}
+
+static GstFlowReturn
+gst_rtp_h264_pay_payload_nal (GstBaseRTPPayload * basepayload, guint8 * data,
+    guint size, GstClockTime timestamp, GstBuffer * buffer_orig)
 {
   GstRtpH264Pay *rtph264pay;
   GstFlowReturn ret;
@@ -558,6 +629,9 @@ gst_rtp_h264_pay_payload_nal (GstBaseRTPPayload * basepayload, guint8 * data,
   guint packet_len, payload_len, mtu;
   GstBuffer *outbuf;
   guint8 *payload;
+  GstBufferList *list = NULL;
+  GstBufferListIterator *it = NULL;
+  gboolean send_spspps;
 
   rtph264pay = GST_RTP_H264_PAY (basepayload);
   mtu = GST_BASE_RTP_PAYLOAD_MTU (rtph264pay);
@@ -565,36 +639,109 @@ gst_rtp_h264_pay_payload_nal (GstBaseRTPPayload * basepayload, guint8 * data,
   nalType = data[0] & 0x1f;
   GST_DEBUG_OBJECT (rtph264pay, "Processing Buffer with NAL TYPE=%d", nalType);
 
+  send_spspps = FALSE;
+
+  /* check if we need to emit an SPS/PPS now */
+  if (nalType == IDR_TYPE_ID && rtph264pay->spspps_interval > 0) {
+    if (rtph264pay->last_spspps != -1) {
+      guint diff;
+
+      /* calculate diff between last SPS/PPS in milliseconds */
+      if (timestamp > rtph264pay->last_spspps)
+        diff = GST_TIME_AS_MSECONDS (timestamp - rtph264pay->last_spspps);
+      else
+        diff = 0;
+
+      GST_DEBUG_OBJECT (rtph264pay, "interval since last SPS/PPS %ums", diff);
+
+      /* bigger than interval, queue SPS/PPS */
+      if (diff >= (rtph264pay->spspps_interval * GST_MSECOND)) {
+        GST_DEBUG_OBJECT (rtph264pay, "time to send SPS/PPS");
+        send_spspps = TRUE;
+      }
+    } else {
+      /* no know previous SPS/PPS time, send now */
+      GST_DEBUG_OBJECT (rtph264pay, "no previous SPS/PPS time, send now");
+      send_spspps = TRUE;
+    }
+  }
+
+  if (send_spspps) {
+    /* we need to send SPS/PPS now first. FIXME, don't use the timestamp for
+     * checking when we need to send SPS/PPS but convert to running_time first. */
+    ret = gst_rtp_h264_pay_send_sps_pps (basepayload, rtph264pay, timestamp);
+    if (ret != GST_FLOW_OK)
+      return ret;
+  }
+
   packet_len = gst_rtp_buffer_calc_packet_len (size, 0, 0);
 
   if (packet_len < mtu) {
     GST_DEBUG_OBJECT (basepayload,
         "NAL Unit fit in one packet datasize=%d mtu=%d", size, mtu);
     /* will fit in one packet */
-    outbuf = gst_rtp_buffer_new_allocate (size, 0, 0);
 
-    GST_BUFFER_TIMESTAMP (outbuf) = timestamp;
+    if (rtph264pay->buffer_list) {
+      /* use buffer lists
+       * first create buffer without payload containing only the RTP header
+       * and then another buffer containing the payload. both buffers will
+       * be then added to the list */
+      outbuf = gst_rtp_buffer_new_allocate (0, 0, 0);
+    } else {
+      /* use the old-fashioned way with a single buffer and memcpy */
+      outbuf = gst_rtp_buffer_new_allocate (size, 0, 0);
+    }
+
     /* only set the marker bit on packets containing access units */
     if (IS_ACCESS_UNIT (nalType)) {
       gst_rtp_buffer_set_marker (outbuf, 1);
     }
 
-    payload = gst_rtp_buffer_get_payload (outbuf);
-    GST_DEBUG_OBJECT (basepayload, "Copying %d bytes to outbuf", size);
-    memcpy (payload, data, size);
+    /* timestamp the outbuffer */
+    GST_BUFFER_TIMESTAMP (outbuf) = timestamp;
 
-    ret = gst_basertppayload_push (basepayload, outbuf);
+    if (rtph264pay->buffer_list) {
+      GstBuffer *paybuf;
+
+      /* create another buffer with the payload. */
+      if (buffer_orig)
+        paybuf = gst_buffer_create_sub (buffer_orig, data -
+            GST_BUFFER_DATA (buffer_orig), size);
+      else {
+        paybuf = gst_buffer_new_and_alloc (size);
+        memcpy (GST_BUFFER_DATA (paybuf), data, size);
+      }
+
+      list = gst_buffer_list_new ();
+      it = gst_buffer_list_iterate (list);
+
+      /* add both buffers to the buffer list */
+      gst_buffer_list_iterator_add_group (it);
+      gst_buffer_list_iterator_add (it, outbuf);
+      gst_buffer_list_iterator_add (it, paybuf);
+
+      gst_buffer_list_iterator_free (it);
+
+      /* push the list to the next element in the pipe */
+      ret = gst_basertppayload_push_list (basepayload, list);
+    } else {
+      payload = gst_rtp_buffer_get_payload (outbuf);
+      GST_DEBUG_OBJECT (basepayload, "Copying %d bytes to outbuf", size);
+      memcpy (payload, data, size);
+
+      ret = gst_basertppayload_push (basepayload, outbuf);
+    }
   } else {
-    /* Fragmentation Units FU-A */
+    /* fragmentation Units FU-A */
     guint8 nalHeader;
     guint limitedSize;
-    int ii = 0, start = 1, end = 0, first = 0;
+    int ii = 0, start = 1, end = 0, pos = 0;
 
     GST_DEBUG_OBJECT (basepayload,
         "NAL Unit DOES NOT fit in one packet datasize=%d mtu=%d", size, mtu);
 
     nalHeader = *data;
-    data++;
+    pos++;
     size--;
 
     ret = GST_FLOW_OK;
@@ -605,13 +752,29 @@ gst_rtp_h264_pay_payload_nal (GstBaseRTPPayload * basepayload, guint8 * data,
     /* We keep 2 bytes for FU indicator and FU Header */
     payload_len = gst_rtp_buffer_calc_payload_len (mtu - 2, 0, 0);
 
+    if (rtph264pay->buffer_list) {
+      list = gst_buffer_list_new ();
+      it = gst_buffer_list_iterate (list);
+    }
+
     while (end == 0) {
       limitedSize = size < payload_len ? size : payload_len;
       GST_DEBUG_OBJECT (basepayload,
           "Inside  FU-A fragmentation limitedSize=%d iteration=%d", limitedSize,
           ii);
 
-      outbuf = gst_rtp_buffer_new_allocate (limitedSize + 2, 0, 0);
+      if (rtph264pay->buffer_list) {
+        /* use buffer lists
+         * first create buffer without payload containing only the RTP header
+         * and then another buffer containing the payload. both buffers will
+         * be then added to the list */
+        outbuf = gst_rtp_buffer_new_allocate (2, 0, 0);
+      } else {
+        /* use the old-fashioned way with a single buffer and memcpy
+         * first create buffer to hold the payload */
+        outbuf = gst_rtp_buffer_new_allocate (limitedSize + 2, 0, 0);
+      }
+
       GST_BUFFER_TIMESTAMP (outbuf) = timestamp;
       payload = gst_rtp_buffer_get_payload (outbuf);
 
@@ -629,20 +792,47 @@ gst_rtp_h264_pay_payload_nal (GstBaseRTPPayload * basepayload, guint8 * data,
       /* FU Header */
       payload[1] = (start << 7) | (end << 6) | (nalHeader & 0x1f);
 
-      memcpy (&payload[2], data + first, limitedSize);
-      GST_DEBUG_OBJECT (basepayload,
-          "recorded %d payload bytes into packet iteration=%d", limitedSize + 2,
-          ii);
+      if (rtph264pay->buffer_list) {
+        GstBuffer *paybuf;
 
-      ret = gst_basertppayload_push (basepayload, outbuf);
-      if (ret != GST_FLOW_OK)
-        break;
+        /* create another buffer to hold the payload */
+        if (buffer_orig)
+          paybuf = gst_buffer_create_sub (buffer_orig, data -
+              GST_BUFFER_DATA (buffer_orig) + pos, limitedSize);
+        else {
+          paybuf = gst_buffer_new_and_alloc (limitedSize);
+          memcpy (GST_BUFFER_DATA (paybuf), data + pos, limitedSize);
+        }
+
+        /* create a new group to hold the header and the payload */
+        gst_buffer_list_iterator_add_group (it);
+
+        /* add both buffers to the buffer list */
+        gst_buffer_list_iterator_add (it, outbuf);
+        gst_buffer_list_iterator_add (it, paybuf);
+
+      } else {
+        memcpy (&payload[2], data + pos, limitedSize);
+        GST_DEBUG_OBJECT (basepayload,
+            "recorded %d payload bytes into packet iteration=%d",
+            limitedSize + 2, ii);
+
+        ret = gst_basertppayload_push (basepayload, outbuf);
+        if (ret != GST_FLOW_OK)
+          break;
+      }
 
       size -= limitedSize;
-      first += limitedSize;
+      pos += limitedSize;
       ii++;
       start = 0;
     }
+
+    if (rtph264pay->buffer_list) {
+      /* free iterator and push the whole buffer list at once */
+      gst_buffer_list_iterator_free (it);
+      ret = gst_basertppayload_push_list (basepayload, list);
+    }
   }
   return ret;
 }
@@ -653,9 +843,10 @@ gst_rtp_h264_pay_handle_buffer (GstBaseRTPPayload * basepayload,
 {
   GstRtpH264Pay *rtph264pay;
   GstFlowReturn ret;
-  guint size, nal_len;
-  guint8 *data;
+  guint size, nal_len, i;
+  guint8 *data, *nal_data;
   GstClockTime timestamp;
+  GArray *nal_queue;
 
   rtph264pay = GST_RTP_H264_PAY (basepayload);
 
@@ -697,7 +888,8 @@ gst_rtp_h264_pay_handle_buffer (GstBaseRTPPayload * basepayload,
       }
 
       ret =
-          gst_rtp_h264_pay_payload_nal (basepayload, data, nal_len, timestamp);
+          gst_rtp_h264_pay_payload_nal (basepayload, data, nal_len, timestamp,
+          buffer);
       if (ret != GST_FLOW_OK)
         break;
 
@@ -706,6 +898,7 @@ gst_rtp_h264_pay_handle_buffer (GstBaseRTPPayload * basepayload,
     }
   } else {
     guint next;
+    gboolean update = FALSE;
 
     /* get offset of first start code */
     next = next_start_code (data, size);
@@ -714,10 +907,16 @@ gst_rtp_h264_pay_handle_buffer (GstBaseRTPPayload * basepayload,
      * will not collect data. */
     data += next;
     size -= next;
+    nal_data = data;
+    nal_queue = rtph264pay->queue;
+
+    /* array must be empty when we get here */
+    g_assert (nal_queue->len == 0);
 
     GST_DEBUG_OBJECT (basepayload, "found first start at %u, bytes left %u",
         next, size);
 
+    /* first pass to locate NALs and parse SPS/PPS */
     while (size > 4) {
       /* skip start code */
       data += 4;
@@ -744,39 +943,71 @@ gst_rtp_h264_pay_handle_buffer (GstBaseRTPPayload * basepayload,
       GST_DEBUG_OBJECT (basepayload, "found next start at %u of size %u", next,
           nal_len);
 
-      if (rtph264pay->profile_level_id != NULL &&
-          rtph264pay->sprop_parameter_sets != NULL) {
+      if (rtph264pay->sprop_parameter_sets != NULL) {
+        /* explicitly set profile and sprop, use those */
         if (rtph264pay->update_caps) {
-          gst_basertppayload_set_outcaps (basepayload, "profile-level-id",
-              G_TYPE_STRING, rtph264pay->profile_level_id,
-              "sprop-parameter-sets", G_TYPE_STRING,
-              rtph264pay->sprop_parameter_sets, NULL);
+          if (!gst_basertppayload_set_outcaps (basepayload,
+                  "sprop-parameter-sets", G_TYPE_STRING,
+                  rtph264pay->sprop_parameter_sets, NULL))
+            goto caps_rejected;
+
           rtph264pay->update_caps = FALSE;
 
-          GST_DEBUG
-              ("outcaps udpate: profile-level-id=%s, sprop-parameter-sets=%s",
-              rtph264pay->profile_level_id, rtph264pay->sprop_parameter_sets);
+          GST_DEBUG ("outcaps udpate: sprop-parameter-sets=%s",
+              rtph264pay->sprop_parameter_sets);
         }
       } else {
         /* We know our stream is a valid H264 NAL packet,
          * go parse it for SPS/PPS to enrich the caps */
-        gst_rtp_h264_pay_parse_sps_pps (basepayload, data, nal_len);
+        /* order: make sure to check nal */
+        update =
+            gst_rtp_h264_pay_decode_nal (rtph264pay, data, nal_len, timestamp)
+            || update;
       }
+      /* move to next NAL packet */
+      data += nal_len;
+      size -= nal_len;
+
+      g_array_append_val (nal_queue, nal_len);
+    }
+
+    /* if has new SPS & PPS, update the output caps */
+    if (G_UNLIKELY (update))
+      if (!gst_rtp_h264_pay_set_sps_pps (basepayload))
+        goto caps_rejected;
+
+    /* second pass to payload and push */
+    data = nal_data;
+    for (i = 0; i < nal_queue->len; i++) {
+      nal_len = g_array_index (nal_queue, guint, i);
+      /* skip start code */
+      data += 4;
 
       /* put the data in one or more RTP packets */
       ret =
-          gst_rtp_h264_pay_payload_nal (basepayload, data, nal_len, timestamp);
-      if (ret != GST_FLOW_OK)
+          gst_rtp_h264_pay_payload_nal (basepayload, data, nal_len, timestamp,
+          buffer);
+      if (ret != GST_FLOW_OK) {
         break;
+      }
 
       /* move to next NAL packet */
       data += nal_len;
       size -= nal_len;
     }
+    g_array_set_size (nal_queue, 0);
   }
+
   gst_buffer_unref (buffer);
 
   return ret;
+
+caps_rejected:
+
+  GST_WARNING_OBJECT (basepayload, "Could not set outcaps");
+  g_array_set_size (nal_queue, 0);
+  gst_buffer_unref (buffer);
+  return GST_FLOW_NOT_NEGOTIATED;
 }
 
 static void
@@ -789,9 +1020,6 @@ gst_rtp_h264_pay_set_property (GObject * object, guint prop_id,
 
   switch (prop_id) {
     case PROP_PROFILE_LEVEL_ID:
-      g_free (rtph264pay->profile_level_id);
-      rtph264pay->profile_level_id = g_value_dup_string (value);
-      rtph264pay->update_caps = TRUE;
       break;
     case PROP_SPROP_PARAMETER_SETS:
       g_free (rtph264pay->sprop_parameter_sets);
@@ -801,6 +1029,12 @@ gst_rtp_h264_pay_set_property (GObject * object, guint prop_id,
     case PROP_SCAN_MODE:
       rtph264pay->scan_mode = g_value_get_enum (value);
       break;
+    case PROP_BUFFER_LIST:
+      rtph264pay->buffer_list = g_value_get_boolean (value);
+      break;
+    case PROP_CONFIG_INTERVAL:
+      rtph264pay->spspps_interval = g_value_get_uint (value);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -817,7 +1051,6 @@ gst_rtp_h264_pay_get_property (GObject * object, guint prop_id,
 
   switch (prop_id) {
     case PROP_PROFILE_LEVEL_ID:
-      g_value_set_string (value, rtph264pay->profile_level_id);
       break;
     case PROP_SPROP_PARAMETER_SETS:
       g_value_set_string (value, rtph264pay->sprop_parameter_sets);
@@ -825,6 +1058,12 @@ gst_rtp_h264_pay_get_property (GObject * object, guint prop_id,
     case PROP_SCAN_MODE:
       g_value_set_enum (value, rtph264pay->scan_mode);
       break;
+    case PROP_BUFFER_LIST:
+      g_value_set_boolean (value, rtph264pay->buffer_list);
+      break;
+    case PROP_CONFIG_INTERVAL:
+      g_value_set_uint (value, rtph264pay->spspps_interval);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
